import re


def toidentifier(string: str) -> str:
    """
    Convert string to a python identifier.

    :param string: The string to convert to a valid python identifier.

    :return: the converted python identifier.
    """
    return re.sub(r"\W|^(?=\d)", "_", string)



import ast

import pendulum
from refactor import ReplacementAction, Rule
from refactor.context import Ancestry, Scope

from airflint.actions import AddNewCall, AddNewImport
from airflint.representatives import DecoratorFinder, ImportFinder
from airflint.utils import toidentifier


class _AddDagDecoratorImport(Rule):
    context_providers = (Scope, DecoratorFinder, ImportFinder)

    def match(self, node: ast.AST) -> AddNewImport:
        assert (
            isinstance(node, ast.ImportFrom)
            and node.module == "airflow"
            and any(alias.name == "DAG" for alias in node.names)
        )
        node_scope = self.context["scope"].resolve(node)
        assert self.context["decorator_finder"].collect("dag", scope=node_scope)
        assert not self.context["import_finder"].collect("dag", scope=node_scope)

        return AddNewImport(node, module="airflow.decorators", names=["dag"])


class _ReplaceDagContextManagerByDagDecorator(Rule):
    def match(self, node):
        assert isinstance(node, ast.With)
        assert (
            DAG := next(
                (
                    item.context_expr
                    for item in node.items
                    if isinstance(item.context_expr, ast.Call)
                    and isinstance(item.context_expr.func, ast.Name)
                    and item.context_expr.func.id == "DAG"
                    and isinstance(item.context_expr.func.ctx, ast.Load)
                ),
                None,
            )
        )
        assert (
            dag_id := next(
                (
                    keyword.value.value
                    for keyword in DAG.keywords
                    if keyword.arg == "dag_id"
                    and isinstance(keyword.value, ast.Constant)
                ),
                "",
            )
        )

        return ReplacementAction(
            node,
            target=ast.FunctionDef(
                name=toidentifier(dag_id),
                args=ast.arguments(
                    posonlyargs=[],
                    args=[],
                    kwonlyargs=[],
                    kw_defaults=[],
                    defaults=[],
                ),
                body=node.body,
                decorator_list=[
                    ast.Call(
                        func=ast.Name(id="dag", ctx=ast.Load()),
                        args=[],
                        keywords=[
                            keyword for keyword in DAG.keywords if keyword.arg != "dag"
                        ],
                    ),
                ],
                lineno=node.lineno,
            ),
        )


class _AddDagCall(Rule):
    def match(self, node: ast.AST) -> AddNewCall:
        assert isinstance(node, ast.FunctionDef)
        assert any(
            isinstance(decorator, ast.Call)
            and isinstance(decorator.func, ast.Name)
            and decorator.func.id == "dag"
            for decorator in node.decorator_list
        )
        assert not (
            isinstance(last_stmt := self.context.tree.body[-1], ast.Expr)
            and isinstance(call := last_stmt.value, ast.Call)
            and isinstance(call.func, ast.Name)
            and call.func.id == node.name
        )

        return AddNewCall(last_stmt, node.name)


EnforceTaskFlowApi = [
    _ReplaceDagContextManagerByDagDecorator,
    _AddDagDecoratorImport,
    _AddDagCall,
]


class _AddDatetimeImport(Rule):
    context_providers = (Scope, ImportFinder)

    def match(self, node: ast.AST) -> AddNewImport:
        assert (
            isinstance(node, ast.ImportFrom)
            and node.module == "airflow.utils.dates"
            and any(alias.name == "days_ago" for alias in node.names)
        )
        assert not self.context["import_finder"].collect(
            "datetime",
            scope=self.context["scope"].resolve(node),
        )

        return AddNewImport(node, module="pendulum", names=["datetime"])


class _ReplaceStartDateDaysAgoByDatetime(Rule):
    context_providers = (Ancestry,)

    def match(self, node: ast.AST) -> ReplacementAction:
        assert (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and isinstance(node.func.ctx, ast.Load)
            and node.func.id == "days_ago"
            and isinstance(days_ago := node.args[0], ast.Constant)
        )
        parents = self.context["ancestry"].get_parents(node)
        assert (
            isinstance(dag_keyword := next(parents), ast.keyword)
            and dag_keyword.arg == "start_date"
        )
        assert (
            isinstance(dag := next(parents), ast.Call)
            and isinstance(dag.func, ast.Name)
            and isinstance(dag.func.ctx, ast.Load)
            and dag.func.id in ["DAG", "dag"]
        )
        assert any(
            keyword.arg == "schedule_interval"
            and isinstance(keyword.value, ast.Constant)
            and keyword.value.value
            for keyword in dag.keywords
        )

        datetime = pendulum.today(tz=pendulum.UTC).subtract(days=days_ago.value)
        return ReplacementAction(
            node,
            target=ast.Call(
                func=ast.Name(id="datetime", ctx=ast.Load()),
                args=[
                    ast.Constant(value=datetime.year),
                    ast.Constant(value=datetime.month),
                    ast.Constant(value=datetime.day),
                ],
                keywords=node.keywords,
            ),
        )


EnforceStaticStartDate = [
    _AddDatetimeImport,
    _ReplaceStartDateDaysAgoByDatetime,
]



from refactor import run

from airflint.rules import dag, task, variable


def main():
    rules = []
    rules.extend(dag.EnforceStaticStartDate)
    rules.extend(dag.EnforceTaskFlowApi)
    rules.extend(task.EnforceTaskFlowApi)
    rules.append(variable.ReplaceVariableGetByJinja)
    run(rules=rules)


if __name__ == "__main__":
    main()



import ast
from dataclasses import dataclass

from refactor import NewStatementAction


@dataclass
class AddNewImport(NewStatementAction):
    """
    Add a new From-Import statement.

    :param module: The import module.
    :param names: The list of names used in alias'.
    """

    module: str
    names: list[str]

    def build(self) -> ast.AST:
        return ast.ImportFrom(
            level=0,
            module=self.module,
            names=[ast.alias(name) for name in self.names],
        )


@dataclass
class AddNewCall(NewStatementAction):
    """
    Add a new Call.

    :param name: The name of the Call to add.
    """

    name: str

    def build(self) -> ast.AST:
        return ast.Call(
            ast.Name(self.name, ast.Load()),
            args=[],
            keywords=[],
        )



import ast
from typing import Optional

from refactor import Representative
from refactor.context import ScopeInfo


class ImportFinder(Representative):
    """Find import by alias name."""

    def collect(self, name: str, scope: ScopeInfo) -> Optional[ast.ImportFrom]:
        return next(
            (
                node
                for node in ast.walk(self.context.tree)
                if isinstance(node, ast.ImportFrom)
                and any(alias.name == name for alias in node.names)
                and scope.can_reach(self.context["scope"].resolve(node))
            ),
            None,
        )


class PythonCallableFinder(Representative):
    """Find python callables references in PythonOperator."""

    def collect(self, name: str, scope: ScopeInfo) -> Optional[ast.Call]:
        return next(
            (
                node
                for node in ast.walk(self.context.tree)
                if isinstance(node, ast.Call)
                and isinstance(node.func, ast.Name)
                and node.func.id in ["PythonOperator", "PythonVirtualenvOperator"]
                and any(
                    keyword.arg == "python_callable"
                    and isinstance(keyword.value, ast.Name)
                    and keyword.value.id == name
                    for keyword in node.keywords
                )
                and scope.can_reach(self.context["scope"].resolve(node))
            ),
            None,
        )


class DecoratorFinder(Representative):
    """Find decorator with the given name."""

    def collect(self, name: str, scope: ScopeInfo) -> Optional[ast.FunctionDef]:
        return next(
            (
                node
                for node in ast.walk(self.context.tree)
                if isinstance(node, ast.FunctionDef)
                and any(
                    isinstance(decorator, ast.Call)
                    and isinstance(decorator.func, ast.Name)
                    and decorator.func.id == name
                    for decorator in node.decorator_list
                )
                and scope.can_reach(self.context["scope"].resolve(node))
            ),
            None,
        )


class NameFinder(Representative):
    """Find decorator with the given name."""

    def collect(self, name: str, scope: ScopeInfo) -> Optional[ast.Name]:
        return next(
            (
                node
                for node in ast.walk(self.context.tree)
                if isinstance(node, ast.Name)
                and node.id == name
                and scope.can_reach(self.context["scope"].resolve(node))
            ),
            None,
        )



import ast
from copy import deepcopy

from refactor import ReplacementAction, Rule
from refactor.context import Scope

from airflint.actions import AddNewImport
from airflint.representatives import ImportFinder, NameFinder, PythonCallableFinder


class _AddTaskDecoratorImport(Rule):
    """Add import for @task decorator."""

    context_providers = (Scope, NameFinder, ImportFinder)

    def match(self, node):
        node_scope = self.context["scope"].resolve(node)
        assert not self.context["name_finder"].collect("task", scope=node_scope)
        assert (
            isinstance(node, ast.ImportFrom)
            and node.module == "airflow.operators.python"
            and any(
                alias.name in ["PythonOperator", "PythonVirtualenvOperator"]
                for alias in node.names
            )
        )
        assert not self.context["import_finder"].collect("task", scope=node_scope)

        return AddNewImport(node, module="airflow.decorators", names=["task"])


class _AddTaskDecorator(Rule):
    """Add @task decorator for python functions to transform them into airflow tasks."""

    context_providers = (Scope, ImportFinder, PythonCallableFinder)

    def match(self, node):
        node_scope = self.context["scope"].resolve(node)
        assert self.context["import_finder"].collect("task", scope=node_scope)
        assert isinstance(node, ast.FunctionDef)
        assert not any(
            decorator.func.id == "task"
            for decorator in node.decorator_list
            if isinstance(decorator, ast.Call) and isinstance(decorator.func, ast.Name)
        ) and not any(
            decorator.func.value.id == "task" and decorator.func.attr == "virtualenv"
            for decorator in node.decorator_list
            if isinstance(decorator, ast.Call)
            and isinstance(decorator.func, ast.Attribute)
            and isinstance(decorator.func.value, ast.Name)
        )
        assert (
            python_operator := self.context["python_callable_finder"].collect(
                node.name,
                scope=node_scope,
            )
        )
        assert all(
            isinstance(keyword.value, ast.Constant)
            for keyword in python_operator.keywords
            if keyword.arg not in ["python_callable", "op_args", "op_kwargs"]
        )
        assert isinstance(python_operator.func, ast.Name)
        TASK_MAPPING = {
            "PythonOperator": ast.Name(id="task", ctx=ast.Load()),
            "PythonVirtualenvOperator": ast.Attribute(
                value=ast.Name(id="task", ctx=ast.Load()),
                attr="virtualenv",
                ctx=ast.Load(),
            ),
        }
        assert (decorator := TASK_MAPPING.get(python_operator.func.id))

        replacement = deepcopy(node)
        replacement.decorator_list.append(
            ast.Call(
                func=decorator,
                args=[],
                keywords=[
                    keyword
                    for keyword in python_operator.keywords
                    if keyword.arg not in ["python_callable", "op_args", "op_kwargs"]
                ],
            ),
        )
        return ReplacementAction(node, replacement)


class _ReplacePythonOperatorByFunctionCall(Rule):
    """Replace PythonOperator calls by function calls which got decorated with the @task decorator."""

    context_providers = (Scope, ImportFinder)

    def match(self, node):
        node_scope = self.context["scope"].resolve(node)
        assert self.context["import_finder"].collect("task", scope=node_scope)
        assert (
            isinstance(node, (ast.Expr, ast.Assign))
            and isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Name)
            and node.value.func.id in ["PythonOperator", "PythonVirtualenvOperator"]
            and isinstance(node.value.func.ctx, ast.Load)
            and any(
                keyword.arg == "python_callable" and isinstance(keyword.value, ast.Name)
                for keyword in node.value.keywords
            )
        )
        assert all(
            isinstance(keyword.value, ast.Constant)
            for keyword in node.value.keywords
            if keyword.arg not in ["python_callable", "op_args", "op_kwargs"]
        )

        args = next(
            (
                keyword.value.elts
                for keyword in node.value.keywords
                if keyword.arg == "op_args"
            ),
            [],
        )
        keywords = next(
            (
                keyword.value.keywords
                for keyword in node.value.keywords
                if keyword.arg == "op_kwargs" and isinstance(keyword.value, ast.Call)
            ),
            next(
                (
                    [
                        ast.keyword(arg=key.value, value=value)
                        for key, value in zip(keyword.value.keys, keyword.value.values)
                        if isinstance(key, ast.Constant)
                    ]
                    for keyword in node.value.keywords
                    if keyword.arg == "op_kwargs"
                    and isinstance(keyword.value, ast.Dict)
                ),
                [],
            ),
        )
        replacement = deepcopy(node)
        replacement.value = ast.Call(
            func=next(
                keyword.value
                for keyword in node.value.keywords
                if keyword.arg == "python_callable"
            ),
            args=args,
            keywords=keywords,
        )
        return ReplacementAction(node, replacement)


EnforceTaskFlowApi = [
    _AddTaskDecoratorImport,
    _AddTaskDecorator,
    _ReplacePythonOperatorByFunctionCall,
]
